@@ -16,19 +16,11 @@ const STYLE =
 const HTML = "<style>" + STYLE + "</style><div></div>";
 const TEST_URL = "data:text/html;charset=utf-8," + encodeURIComponent(HTML);
 
-var highlightedNodeFront, highlighterOptions;
-
 add_task(async function() {
   await addTab(TEST_URL);
   const { inspector, boxmodel } = await openLayoutView();
   await selectNode("div", inspector);
 
-  // Mock the highlighter by replacing the showBoxModel method.
-  inspector.highlighter.showBoxModel = function(nodeFront, options) {
-    highlightedNodeFront = nodeFront;
-    highlighterOptions = options;
-  };
-
   let elt = boxmodel.document.querySelector(".boxmodel-margins");
   await testGuideOnLayoutHover(elt, "margin", inspector);
 
@@ -43,6 +35,11 @@ add_task(async function() {
 });
 
 async function testGuideOnLayoutHover(elt, expectedRegion, inspector) {
+  const { waitForHighlighterTypeShown } = getHighlighterTestHelpers(inspector);
+  const onHighlighterShown = waitForHighlighterTypeShown(
+    inspector.highlighters.TYPES.BOXMODEL
+  );
+
   info("Synthesizing mouseover on the boxmodel-view");
   EventUtils.synthesizeMouse(
     elt,
@@ -53,19 +50,19 @@ async function testGuideOnLayoutHover(elt, expectedRegion, inspector) {
   );
 
   info("Waiting for the node-highlight event from the toolbox");
-  await inspector.highlighter.once("node-highlight");
+  const { nodeFront, options } = await onHighlighterShown;
 
   // Wait for the next event tick to make sure the remaining part of the
   // test is executed after finishing synthesizing mouse event.
   await new Promise(executeSoon);
 
   is(
-    highlightedNodeFront,
+    nodeFront,
     inspector.selection.nodeFront,
     "The right nodeFront was highlighted"
   );
   is(
-    highlighterOptions.region,
+    options.region,
     expectedRegion,
     "Region " + expectedRegion + " was highlighted"
   );
