@@ -8,13 +8,14 @@
 
 #include "mozilla/DebugOnly.h"
 
+#include "ipc/EnumSerializer.h"
 #include "ipc/IPCMessageUtils.h"
 #include "nsExceptionHandler.h"
+#include "nsIHttpChannel.h"
 #include "nsPrintfCString.h"
 #include "nsString.h"
 #include "prio.h"
 #include "mozilla/net/DNS.h"
-#include "TimingStruct.h"
 
 namespace IPC {
 
@@ -25,8 +26,9 @@ struct Permission {
   uint32_t capability, expireType;
   int64_t expireTime;
 
-  Permission() {}
-  Permission(const nsCString& aOrigin, const nsCString& aType,
+  Permission() : capability(0), expireType(0), expireTime(0) {}
+
+  Permission(const nsCString& aOrigin, const nsACString& aType,
              const uint32_t aCapability, const uint32_t aExpireType,
              const int64_t aExpireTime)
       : origin(aOrigin),
@@ -34,6 +36,12 @@ struct Permission {
         capability(aCapability),
         expireType(aExpireType),
         expireTime(aExpireTime) {}
+
+  bool operator==(const Permission& aOther) const {
+    return aOther.origin == origin && aOther.type == type &&
+           aOther.capability == capability && aOther.expireType == expireType &&
+           aOther.expireTime == expireTime;
+  }
 };
 
 template <>
@@ -95,7 +103,7 @@ struct ParamTraits<mozilla::net::NetAddr> {
       if (XRE_IsParentProcess()) {
         nsPrintfCString msg("%d", aParam.raw.family);
         CrashReporter::AnnotateCrashReport(
-            NS_LITERAL_CSTRING("Unknown NetAddr socket family"), msg);
+            CrashReporter::Annotation::UnknownNetAddrSocketFamily, msg);
       }
 
       MOZ_CRASH("Unknown socket family");
@@ -131,52 +139,10 @@ struct ParamTraits<mozilla::net::NetAddr> {
 };
 
 template <>
-struct ParamTraits<mozilla::net::ResourceTimingStruct> {
-  static void Write(Message* aMsg,
-                    const mozilla::net::ResourceTimingStruct& aParam) {
-    WriteParam(aMsg, aParam.domainLookupStart);
-    WriteParam(aMsg, aParam.domainLookupEnd);
-    WriteParam(aMsg, aParam.connectStart);
-    WriteParam(aMsg, aParam.tcpConnectEnd);
-    WriteParam(aMsg, aParam.secureConnectionStart);
-    WriteParam(aMsg, aParam.connectEnd);
-    WriteParam(aMsg, aParam.requestStart);
-    WriteParam(aMsg, aParam.responseStart);
-    WriteParam(aMsg, aParam.responseEnd);
-
-    WriteParam(aMsg, aParam.fetchStart);
-    WriteParam(aMsg, aParam.redirectStart);
-    WriteParam(aMsg, aParam.redirectEnd);
-
-    WriteParam(aMsg, aParam.transferSize);
-    WriteParam(aMsg, aParam.encodedBodySize);
-    WriteParam(aMsg, aParam.protocolVersion);
-
-    WriteParam(aMsg, aParam.cacheReadStart);
-    WriteParam(aMsg, aParam.cacheReadEnd);
-  }
-
-  static bool Read(const Message* aMsg, PickleIterator* aIter,
-                   mozilla::net::ResourceTimingStruct* aResult) {
-    return ReadParam(aMsg, aIter, &aResult->domainLookupStart) &&
-           ReadParam(aMsg, aIter, &aResult->domainLookupEnd) &&
-           ReadParam(aMsg, aIter, &aResult->connectStart) &&
-           ReadParam(aMsg, aIter, &aResult->tcpConnectEnd) &&
-           ReadParam(aMsg, aIter, &aResult->secureConnectionStart) &&
-           ReadParam(aMsg, aIter, &aResult->connectEnd) &&
-           ReadParam(aMsg, aIter, &aResult->requestStart) &&
-           ReadParam(aMsg, aIter, &aResult->responseStart) &&
-           ReadParam(aMsg, aIter, &aResult->responseEnd) &&
-           ReadParam(aMsg, aIter, &aResult->fetchStart) &&
-           ReadParam(aMsg, aIter, &aResult->redirectStart) &&
-           ReadParam(aMsg, aIter, &aResult->redirectEnd) &&
-           ReadParam(aMsg, aIter, &aResult->transferSize) &&
-           ReadParam(aMsg, aIter, &aResult->encodedBodySize) &&
-           ReadParam(aMsg, aIter, &aResult->protocolVersion) &&
-           ReadParam(aMsg, aIter, &aResult->cacheReadStart) &&
-           ReadParam(aMsg, aIter, &aResult->cacheReadEnd);
-  }
-};
+struct ParamTraits<nsIHttpChannel::FlashPluginState>
+    : public ContiguousEnumSerializerInclusive<
+          nsIHttpChannel::FlashPluginState, nsIHttpChannel::FlashPluginUnknown,
+          nsIHttpChannel::FlashPluginLastValue> {};
 
 }  // namespace IPC
 
