@@ -101,7 +101,8 @@ void nsFileCopyEvent::DoCopy() {
     rv = mSource->ReadSegments(NS_CopySegmentToStream, mDest, num, &result);
     if (NS_FAILED(rv)) break;
     if (result != (uint32_t)num) {
-      rv = NS_ERROR_FILE_DISK_FULL;  // stopped prematurely (out of disk space)
+      // stopped prematurely (out of disk space)
+      rv = NS_ERROR_FILE_NO_DEVICE_SPACE;
       break;
     }
 
@@ -305,8 +306,9 @@ nsresult nsFileChannel::MakeFileInputStream(nsIFile* file,
 
   if (isDir) {
     rv = nsDirectoryIndexStream::Create(file, getter_AddRefs(stream));
-    if (NS_SUCCEEDED(rv) && !HasContentTypeHint())
+    if (NS_SUCCEEDED(rv) && !HasContentTypeHint()) {
       contentType.AssignLiteral(APPLICATION_HTTP_INDEX_FORMAT);
+    }
   } else {
     rv = NS_NewLocalFileInputStream(getter_AddRefs(stream), file, -1, -1,
                                     async ? nsIFileInputStream::DEFER_OPEN : 0);
@@ -334,12 +336,12 @@ nsresult nsFileChannel::OpenContentStream(bool async, nsIInputStream** result,
   if (NS_FAILED(rv)) return rv;
 
   nsCOMPtr<nsIURI> newURI;
-  rv = fileHandler->ReadURLFile(file, getter_AddRefs(newURI));
-  if (NS_SUCCEEDED(rv)) {
+  if (NS_SUCCEEDED(fileHandler->ReadURLFile(file, getter_AddRefs(newURI))) ||
+      NS_SUCCEEDED(fileHandler->ReadShellLink(file, getter_AddRefs(newURI)))) {
     nsCOMPtr<nsIChannel> newChannel;
     rv = NS_NewChannel(getter_AddRefs(newChannel), newURI,
                        nsContentUtils::GetSystemPrincipal(),
-                       nsILoadInfo::SEC_ALLOW_CROSS_ORIGIN_DATA_IS_NULL,
+                       nsILoadInfo::SEC_ALLOW_CROSS_ORIGIN_SEC_CONTEXT_IS_NULL,
                        nsIContentPolicy::TYPE_OTHER);
 
     if (NS_FAILED(rv)) return rv;
@@ -376,8 +378,9 @@ nsresult nsFileChannel::OpenContentStream(bool async, nsIInputStream** result,
     // to something other than "unknown" to avoid triggering the content-type
     // sniffer code in nsBaseChannel.
     // However, don't override explicitly set types.
-    if (!HasContentTypeHint())
-      SetContentType(NS_LITERAL_CSTRING(APPLICATION_OCTET_STREAM));
+    if (!HasContentTypeHint()) {
+      SetContentType(nsLiteralCString(APPLICATION_OCTET_STREAM));
+    }
   } else {
     nsAutoCString contentType;
     rv = MakeFileInputStream(file, stream, contentType, async);
