@@ -8,19 +8,19 @@
 #ifndef mozilla_net_HttpChannelParent_h
 #define mozilla_net_HttpChannelParent_h
 
-#include "ADivertableParentChannel.h"
 #include "nsHttp.h"
 #include "mozilla/net/PHttpChannelParent.h"
 #include "mozilla/net/NeckoCommon.h"
 #include "mozilla/net/NeckoParent.h"
 #include "mozilla/MozPromise.h"
-#include "nsIObserver.h"
 #include "nsIParentRedirectingChannel.h"
 #include "nsIProgressEventSink.h"
+#include "nsIChannelEventSink.h"
+#include "nsIRedirectResultListener.h"
 #include "nsHttpChannel.h"
-#include "nsIAuthPromptProvider.h"
 #include "mozilla/dom/ipc/IdType.h"
 #include "nsIDeprecationWarner.h"
+#include "nsIMultiPartChannel.h"
 
 class nsICacheEntry;
 
@@ -35,13 +35,12 @@ namespace mozilla {
 
 namespace dom {
 class BrowserParent;
-class PBrowserOrId;
 }  // namespace dom
 
 namespace net {
 
 class HttpBackgroundChannelParent;
-class HttpChannelParentListener;
+class ParentChannelListener;
 class ChannelEventQueue;
 
 // Note: nsIInterfaceRequestor must be the first base so that do_QueryObject()
@@ -52,11 +51,12 @@ class HttpChannelParent final : public nsIInterfaceRequestor,
                                 public PHttpChannelParent,
                                 public nsIParentRedirectingChannel,
                                 public nsIProgressEventSink,
-                                public ADivertableParentChannel,
-                                public nsIAuthPromptProvider,
                                 public nsIDeprecationWarner,
                                 public HttpChannelSecurityWarningReporter,
-                                public nsIAsyncVerifyRedirectReadyCallback {
+                                public nsIAsyncVerifyRedirectReadyCallback,
+                                public nsIChannelEventSink,
+                                public nsIRedirectResultListener,
+                                public nsIMultiPartChannelListener {
   virtual ~HttpChannelParent();
 
  public:
@@ -67,32 +67,19 @@ class HttpChannelParent final : public nsIInterfaceRequestor,
   NS_DECL_NSIPARENTREDIRECTINGCHANNEL
   NS_DECL_NSIPROGRESSEVENTSINK
   NS_DECL_NSIINTERFACEREQUESTOR
-  NS_DECL_NSIAUTHPROMPTPROVIDER
   NS_DECL_NSIDEPRECATIONWARNER
   NS_DECL_NSIASYNCVERIFYREDIRECTREADYCALLBACK
+  NS_DECL_NSICHANNELEVENTSINK
+  NS_DECL_NSIREDIRECTRESULTLISTENER
+  NS_DECL_NSIMULTIPARTCHANNELLISTENER
 
   NS_DECLARE_STATIC_IID_ACCESSOR(HTTP_CHANNEL_PARENT_IID)
 
-  HttpChannelParent(const dom::PBrowserOrId& iframeEmbedding,
-                    nsILoadContext* aLoadContext, PBOverrideStatus aStatus);
+  HttpChannelParent(dom::BrowserParent* iframeEmbedding,
+                    nsILoadContext* aLoadContext,
+                    PBOverrideStatus aOverrideStatus);
 
-  MOZ_MUST_USE bool Init(const HttpChannelCreationArgs& aOpenArgs);
-
-  // ADivertableParentChannel functions.
-  void DivertTo(nsIStreamListener* aListener) override;
-  MOZ_MUST_USE nsresult SuspendForDiversion() override;
-  MOZ_MUST_USE nsresult SuspendMessageDiversion() override;
-  MOZ_MUST_USE nsresult ResumeMessageDiversion() override;
-  MOZ_MUST_USE nsresult CancelDiversion() override;
-
-  // Calls OnStartRequest for "DivertTo" listener, then notifies child channel
-  // that it should divert OnDataAvailable and OnStopRequest calls to this
-  // parent channel.
-  void StartDiversion();
-
-  // Handles calling OnStart/Stop if there are errors during diversion.
-  // Called asynchronously from FailDiversion.
-  void NotifyDiversionFailed(nsresult aErrorCode);
+  [[nodiscard]] bool Init(const HttpChannelCreationArgs& aArgs);
 
   // Forwarded to nsHttpChannel::SetApplyConversion.
   void SetApplyConversion(bool aApplyConversion) {
@@ -101,9 +88,9 @@ class HttpChannelParent final : public nsIInterfaceRequestor,
     }
   }
 
-  MOZ_MUST_USE nsresult
-  OpenAlternativeOutputStream(const nsACString& type, int64_t predictedSize,
-                              nsIAsyncOutputStream** _retval);
+  [[nodiscard]] nsresult OpenAlternativeOutputStream(
+      const nsACString& type, int64_t predictedSize,
+      nsIAsyncOutputStream** _retval);
 
   // Callbacks for each asynchronous tasks required in AsyncOpen
   // procedure, will call InvokeAsyncOpen when all the expected
@@ -121,26 +108,36 @@ class HttpChannelParent final : public nsIInterfaceRequestor,
   // Callback while background channel is destroyed.
   void OnBackgroundParentDestroyed();
 
-  base::ProcessId OtherPid() const override;
+  base::ProcessId OtherPid() const;
 
-  // Calling this method will cancel the HttpChannelChild because the consumer
-  // needs to be relocated to another process.
-  // Any OnStart/Stop/DataAvailable calls that follow will not be sent to the
-  // child channel.
-  void CancelChildCrossProcessRedirect();
+  // Inform the child actor that our referrer info was modified late during
+  // BeginConnect.
+  void OverrideReferrerInfoDuringBeginConnect(nsIReferrerInfo* aReferrerInfo);
 
-  already_AddRefed<HttpChannelParentListener> GetParentListener();
+  // Set the cookie string, which will be informed to the child actor during
+  // PHttpBackgroundChannel::OnStartRequest. Note that CookieService also sends
+  // the information to all actors via PContent, a main thread IPC, which could
+  // be slower than background IPC PHttpBackgroundChannel::OnStartRequest.
+  // Therefore, another cookie notification via PBackground is needed to
+  // guarantee the listener in child has the necessary cookies before
+  // OnStartRequest.
+  void SetCookie(nsCString&& aCookie);
+
+  using ChildEndpointPromise =
+      MozPromise<ipc::Endpoint<extensions::PStreamFilterChild>, bool, true>;
+  [[nodiscard]] RefPtr<ChildEndpointPromise> AttachStreamFilter(
+      Endpoint<extensions::PStreamFilterParent>&& aParentEndpoint,
+      Endpoint<extensions::PStreamFilterChild>&& aChildEndpoint);
 
  protected:
   // used to connect redirected-to channel in parent with just created
   // ChildChannel.  Used during redirects.
-  MOZ_MUST_USE bool ConnectChannel(const uint32_t& channelId,
-                                   const bool& shouldIntercept);
+  [[nodiscard]] bool ConnectChannel(const uint32_t& registrarId);
 
-  MOZ_MUST_USE bool DoAsyncOpen(
+  [[nodiscard]] bool DoAsyncOpen(
       const URIParams& uri, const Maybe<URIParams>& originalUri,
       const Maybe<URIParams>& docUri, nsIReferrerInfo* aReferrerInfo,
-      const Maybe<URIParams>& internalRedirectUri,
+      const Maybe<URIParams>& aAPIRedirectToURI,
       const Maybe<URIParams>& topWindowUri, const uint32_t& loadFlags,
       const RequestHeaderTuples& requestHeaders, const nsCString& requestMethod,
       const Maybe<IPCStream>& uploadStream, const bool& uploadStreamHasHeaders,
@@ -148,23 +145,20 @@ class HttpChannelParent final : public nsIInterfaceRequestor,
       const uint8_t& redirectionLimit, const bool& allowSTS,
       const uint32_t& thirdPartyFlags, const bool& doResumeAt,
       const uint64_t& startPos, const nsCString& entityID,
-      const bool& chooseApplicationCache, const nsCString& appCacheClientID,
-      const bool& allowSpdy, const bool& allowAltSvc,
+      const bool& allowSpdy, const bool& allowHttp3, const bool& allowAltSvc,
       const bool& beConservative, const uint32_t& tlsFlags,
-      const Maybe<LoadInfoArgs>& aLoadInfoArgs,
-      const Maybe<nsHttpResponseHead>& aSynthesizedResponseHead,
-      const nsCString& aSecurityInfoSerialization, const uint32_t& aCacheKey,
+      const Maybe<LoadInfoArgs>& aLoadInfoArgs, const uint32_t& aCacheKey,
       const uint64_t& aRequestContextID,
       const Maybe<CorsPreflightArgs>& aCorsPreflightArgs,
       const uint32_t& aInitialRwin, const bool& aBlockAuthPrompt,
-      const bool& aSuspendAfterSynthesizeResponse,
-      const bool& aAllowStaleCacheContent, const nsCString& aContentTypeHint,
+      const bool& aAllowStaleCacheContent,
+      const bool& aPreferCacheLoadOverBypass, const nsCString& aContentTypeHint,
       const uint32_t& aCorsMode, const uint32_t& aRedirectMode,
       const uint64_t& aChannelId, const nsString& aIntegrityMetadata,
       const uint64_t& aContentWindowId,
       const nsTArray<PreferredAlternativeDataTypeParams>&
           aPreferredAlternativeTypes,
-      const uint64_t& aTopLevelOuterContentWindowId,
+      const uint64_t& aTopBrowsingContextId,
       const TimeStamp& aLaunchServiceWorkerStart,
       const TimeStamp& aLaunchServiceWorkerEnd,
       const TimeStamp& aDispatchFetchEventStart,
@@ -178,48 +172,33 @@ class HttpChannelParent final : public nsIInterfaceRequestor,
       const int16_t& priority) override;
   virtual mozilla::ipc::IPCResult RecvSetClassOfService(
       const uint32_t& cos) override;
-  virtual mozilla::ipc::IPCResult RecvSetCacheTokenCachedCharset(
-      const nsCString& charset) override;
   virtual mozilla::ipc::IPCResult RecvSuspend() override;
   virtual mozilla::ipc::IPCResult RecvResume() override;
-  virtual mozilla::ipc::IPCResult RecvCancel(const nsresult& status) override;
+  virtual mozilla::ipc::IPCResult RecvCancel(
+      const nsresult& status, const uint32_t& requestBlockingReason) override;
   virtual mozilla::ipc::IPCResult RecvRedirect2Verify(
       const nsresult& result, const RequestHeaderTuples& changedHeaders,
-      const ChildLoadInfoForwarderArgs& aLoadInfoForwarder,
+      const uint32_t& aSourceRequestBlockingReason,
+      const Maybe<ChildLoadInfoForwarderArgs>& aTargetLoadInfoForwarder,
       const uint32_t& loadFlags, nsIReferrerInfo* aReferrerInfo,
       const Maybe<URIParams>& apiRedirectUri,
-      const Maybe<CorsPreflightArgs>& aCorsPreflightArgs,
-      const bool& aChooseAppcache) override;
+      const Maybe<CorsPreflightArgs>& aCorsPreflightArgs) override;
   virtual mozilla::ipc::IPCResult RecvDocumentChannelCleanup(
       const bool& clearCacheEntry) override;
-  virtual mozilla::ipc::IPCResult RecvMarkOfflineCacheEntryAsForeign() override;
-  virtual mozilla::ipc::IPCResult RecvDivertOnDataAvailable(
-      const nsCString& data, const uint64_t& offset,
-      const uint32_t& count) override;
-  virtual mozilla::ipc::IPCResult RecvDivertOnStopRequest(
-      const nsresult& statusCode) override;
-  virtual mozilla::ipc::IPCResult RecvDivertComplete() override;
-  virtual mozilla::ipc::IPCResult RecvCrossProcessRedirectDone(
-      const nsresult& aResult) override;
   virtual mozilla::ipc::IPCResult RecvRemoveCorsPreflightCacheEntry(
       const URIParams& uri,
-      const mozilla::ipc::PrincipalInfo& requestingPrincipal) override;
+      const mozilla::ipc::PrincipalInfo& requestingPrincipal,
+      const OriginAttributes& originAttributes) override;
   virtual mozilla::ipc::IPCResult RecvBytesRead(const int32_t& aCount) override;
   virtual mozilla::ipc::IPCResult RecvOpenOriginalCacheInputStream() override;
   virtual mozilla::ipc::IPCResult RecvOpenAltDataCacheInputStream(
       const nsCString& aType) override;
   virtual void ActorDestroy(ActorDestroyReason why) override;
 
-  // Supporting function for ADivertableParentChannel.
-  MOZ_MUST_USE nsresult ResumeForDiversion();
-
-  // Asynchronously calls NotifyDiversionFailed.
-  void FailDiversion(nsresult aErrorCode);
-
-  friend class HttpChannelParentListener;
+  friend class ParentChannelListener;
   RefPtr<mozilla::dom::BrowserParent> mBrowserParent;
 
-  MOZ_MUST_USE nsresult ReportSecurityMessage(
+  [[nodiscard]] nsresult ReportSecurityMessage(
       const nsAString& aMessageTag, const nsAString& aMessageCategory) override;
   nsresult LogBlockedCORSRequest(const nsAString& aMessage,
                                  const nsACString& aCategory) override;
@@ -229,21 +208,13 @@ class HttpChannelParent final : public nsIInterfaceRequestor,
 
   // Calls SendDeleteSelf and sets mIPCClosed to true because we should not
   // send any more messages after that. Bug 1274886
-  MOZ_MUST_USE bool DoSendDeleteSelf();
+  [[nodiscard]] bool DoSendDeleteSelf();
   // Called to notify the parent channel to not send any more IPC messages.
   virtual mozilla::ipc::IPCResult RecvDeletingChannel() override;
-  virtual mozilla::ipc::IPCResult RecvFinishInterceptedRedirect() override;
 
  private:
   void UpdateAndSerializeSecurityInfo(nsACString& aSerializedSecurityInfoOut);
 
-  void DivertOnDataAvailable(const nsCString& data, const uint64_t& offset,
-                             const uint32_t& count);
-  void DivertOnStopRequest(const nsresult& statusCode);
-  void DivertComplete();
-  void MaybeFlushPendingDiversion();
-  void ResponseSynthesized();
-
   // final step for Redirect2Verify procedure, will be invoked while both
   // redirecting and redirected channel are ready or any error happened.
   // OnRedirectVerifyCallback will be invoked for finishing the async
@@ -256,7 +227,7 @@ class HttpChannelParent final : public nsIInterfaceRequestor,
   // id, a promise will be returned so the caller can append callbacks on it.
   // If called multiple times before mBgParent is available, the same promise
   // will be returned and the callbacks will be invoked in order.
-  MOZ_MUST_USE RefPtr<GenericNonExclusivePromise> WaitForBgParent();
+  [[nodiscard]] RefPtr<GenericNonExclusivePromise> WaitForBgParent();
 
   // Remove the association with background channel after main-thread IPC
   // is about to be destroyed or no further event is going to be sent, i.e.,
@@ -270,9 +241,6 @@ class HttpChannelParent final : public nsIInterfaceRequestor,
   int32_t mSendWindowSize;
 
   friend class HttpBackgroundChannelParent;
-  friend class DivertDataAvailableEvent;
-  friend class DivertStopRequestEvent;
-  friend class DivertCompleteEvent;
 
   RefPtr<HttpBaseChannel> mChannel;
   nsCOMPtr<nsICacheEntry> mCacheEntry;
@@ -280,15 +248,10 @@ class HttpChannelParent final : public nsIInterfaceRequestor,
   nsCOMPtr<nsIChannel> mRedirectChannel;
   nsCOMPtr<nsIAsyncVerifyRedirectCallback> mRedirectCallback;
 
-  nsAutoPtr<class nsHttpChannel::OfflineCacheEntryAsForeignMarker>
-      mOfflineForeignMarker;
   nsCOMPtr<nsILoadContext> mLoadContext;
   RefPtr<nsHttpHandler> mHttpHandler;
 
-  RefPtr<HttpChannelParentListener> mParentListener;
-  // The listener we are diverting to or will divert to if mPendingDiversion
-  // is set.
-  nsCOMPtr<nsIStreamListener> mDivertListener;
+  RefPtr<ParentChannelListener> mParentListener;
 
   RefPtr<ChannelEventQueue> mEventQ;
 
@@ -297,8 +260,6 @@ class HttpChannelParent final : public nsIInterfaceRequestor,
   MozPromiseHolder<GenericNonExclusivePromise> mPromise;
   MozPromiseRequestHolder<GenericNonExclusivePromise> mRequest;
 
-  dom::TabId mNestedFrameId;
-
   // To calculate the delay caused by the e10s back-pressure suspension
   TimeStamp mResumedTimestamp;
 
@@ -306,13 +267,21 @@ class HttpChannelParent final : public nsIInterfaceRequestor,
 
   // Corresponding redirect channel registrar Id. 0 means redirection is not
   // started.
-  uint32_t mRedirectRegistrarId = 0;
+  uint64_t mRedirectChannelId = 0;
 
   PBOverrideStatus mPBOverride;
 
   // Set to the canceled status value if the main channel was canceled.
   nsresult mStatus;
 
+  // The referrer info, set during nsHttpChannel::BeginConnect, to override the
+  // original one. This info will be sent in OnStartRequest.
+  nsCOMPtr<nsIReferrerInfo> mOverrideReferrerInfo;
+
+  // The cookie string in Set-Cookie header. This info will be sent in
+  // OnStartRequest.
+  nsCString mCookie;
+
   // OnStatus is always called before OnProgress.
   // Set true in OnStatus if next OnProgress can be ignored
   // since the information can be recontructed from ODA.
@@ -322,38 +291,24 @@ class HttpChannelParent final : public nsIInterfaceRequestor,
   uint8_t mReceivedRedirect2Verify : 1;
   uint8_t mHasSuspendedByBackPressure : 1;
 
-  // Indicates that diversion has been requested, but we could not start it
-  // yet because the channel is still being opened with a synthesized response.
-  uint8_t mPendingDiversion : 1;
-  // Once set, no OnStart/OnData/OnStop calls should be accepted; conversely, it
-  // must be set when RecvDivertOnData/~DivertOnStop/~DivertComplete are
-  // received from the child channel.
-  uint8_t mDivertingFromChild : 1;
-
-  // Set if OnStart|StopRequest was called during a diversion from the child.
-  uint8_t mDivertedOnStartRequest : 1;
-
-  uint8_t mSuspendedForDiversion : 1;
-
-  // Set if this channel should be suspended after synthesizing a response.
-  uint8_t mSuspendAfterSynthesizeResponse : 1;
-  // Set if this channel will synthesize its response.
-  uint8_t mWillSynthesizeResponse : 1;
-
   // Set if we get the result of and cache |mNeedFlowControl|
   uint8_t mCacheNeedFlowControlInitialized : 1;
   uint8_t mNeedFlowControl : 1;
   uint8_t mSuspendedForFlowControl : 1;
 
-  // The child channel was cancelled, as the consumer was relocated to another
-  // process.
-  uint8_t mDoingCrossProcessRedirect : 1;
+  // Defaults to false. Is set to true at the begining of OnStartRequest.
+  // Used to ensure methods can't be called before OnStartRequest.
+  uint8_t mAfterOnStartRequestBegun : 1;
 
   // Number of events to wait before actually invoking AsyncOpen on the main
   // channel. For each asynchronous step required before InvokeAsyncOpen, should
   // increase 1 to mAsyncOpenBarrier and invoke TryInvokeAsyncOpen after
   // finished. This attribute is main thread only.
   uint8_t mAsyncOpenBarrier = 0;
+
+  // When true, ODAs are sent from the socket process to the child process
+  // directly.
+  uint8_t mDataSentToChildProcess : 1;
 };
 
 NS_DEFINE_STATIC_IID_ACCESSOR(HttpChannelParent, HTTP_CHANNEL_PARENT_IID)
