@@ -50,6 +50,13 @@ add_test(function test_dns() {
 });
 
 add_test(function test_sockets() {
+  // TODO: enable this test in bug 1581892.
+  if (mozinfo.socketprocess_networking) {
+    info("skip test_sockets");
+    run_next_test();
+    return;
+  }
+
   let sts = Cc["@mozilla.org/network/socket-transport-service;1"].getService(
     Ci.nsISocketTransportService
   );
@@ -59,6 +66,7 @@ add_test(function test_sockets() {
     [],
     "127.0.0.1",
     gServerSocket.port,
+    null,
     null
   );
   let listener = {
@@ -91,6 +99,9 @@ function run_test() {
     true
   );
 
+  // We always resolve localhost as it's hardcoded without the following pref:
+  Services.prefs.setBoolPref("network.proxy.allow_hijacking_localhost", true);
+
   let ioService = Cc["@mozilla.org/network/io-service;1"].getService(
     Ci.nsIIOService
   );
@@ -105,6 +116,7 @@ function run_test() {
   channel.open();
 
   gServerSocket.init(-1, true, -1);
+  Services.prefs.clearUserPref("network.proxy.allow_hijacking_localhost");
 
   run_next_test();
 }
